@@ -23,6 +23,12 @@
 
 from .mail import KryptoMIME
 
+class KeyMissingError(Exception):
+    def __init__(self, key):
+        self.key = key
+    def __str__(self):
+        return "key missing: "+self.key
+
 class PGPMIME(KryptoMIME):
 
     def find_key(self,addr,secret=False):
@@ -67,7 +73,6 @@ class PGPMIME(KryptoMIME):
     @staticmethod
     def _plaintext(mail, inline = False, protect=False):
         # Extract / generate plaintext
-        from email.parser import HeaderParser
         import copy
         multipart = mail.is_multipart()
         if not multipart and inline: return mail.get_payload(), None
@@ -86,7 +91,8 @@ class PGPMIME(KryptoMIME):
         ciphertext = None
         is_pgpmime = False
         # Check: Is inline pgp?
-        if type(mail)==str: mail = Parser().parsestr(mail)
+        from .mail import protect_mail
+        if type(mail)==str: mail = protect_mail(mail,ending=None)
         if mail.get_content_type()=='application/pgp' or mail.get_param('x-action')=='pgp-encrypted':
             ciphertext = mail.get_payload()
             is_pgpmime = False
@@ -131,15 +137,16 @@ class PGPMIME(KryptoMIME):
 
     @staticmethod
     def _signature(mail):
-        from email.parser import HeaderParser, Parser
-        from .mail import as_protected, _mail_raw, _mail_transfer_content
+        from .mail import as_protected, _mail_raw, _mail_transfer_content, protect_mail
+        from sys import version_info as vs
         payload = ''
         signatures = []
         rawmail = mail
         mail = as_protected(mail)
+        fix_nl = (vs[0]==2 and vs[1]*10+vs[2]<77) or (vs[0]==3 and vs[1]*10+vs[2]<35)
         if mail.is_multipart():
             rawmail = as_protected(mail,headersonly=True)
-            rawmail.epilogue='' # ensure final newline, workaround for http://bugs.python.org/issue14983
+            if fix_nl: rawmail.epilogue='' # ensure final newline, workaround for http://bugs.python.org/issue14983
             if mail.get_content_type()=='multipart/signed' and mail.get_param('protocol')=='application/pgp-signature':
                 # handle detached signatures, these look like:
                 for submsg in mail.get_payload():
@@ -202,18 +209,18 @@ class PGPMIME(KryptoMIME):
                 if line.rstrip()=='-----BEGIN PGP SIGNATURE-----': break
                 text += line
             signatures = [None]
-            rawmail = HeaderParser().parsestr(mail.as_string())
-            rawmail.del_param("x-action")
-            rawmail.set_payload(text)
+            mail.del_param("x-action")
+            mail.set_payload(text.rstrip()) # remove last line ending
+            rawmail = mail
         return payload, signatures, rawmail
 
     @staticmethod
     def _decoded(mail,plaintext,is_pgpmime):
         # Check transfer type
-        from email.parser import HeaderParser, Parser
         from .mail import _mail_addreplace_header, _mail_transfer_content
-        mail = HeaderParser().parsestr(mail.as_string())
-        tmpmsg = Parser().parsestr(plaintext)
+        from .mail import protect_mail
+        mail = protect_mail(mail,ending=None)
+        tmpmsg = protect_mail(plaintext,ending=None)
         if mail.get_content_type()=='application/pgp': mail.set_type("text/plain")
         mail.del_param("x-action")
         if tmpmsg.is_multipart() and len(tmpmsg.get_payload())==1:
@@ -237,9 +244,9 @@ class PGPMIME(KryptoMIME):
         :returns: Whether the email is encrypted and whether it is signed (if it is not encrypted).
         :rtype: (bool,bool/None)
         """
-        from email.parser import Parser
         from email.message import Message
-        if type(mail)==str: mail = Parser().parsestr(mail)
+        from .mail import protect_mail
+        if type(mail)==str: mail = protect_mail(mail,ending=None)
         elif not isinstance(mail,Message): return False, False
         ciphertext, is_pgpmime = self._ciphertext(mail)
         if not ciphertext is None: return True, None
@@ -253,10 +260,11 @@ class PGPMIME(KryptoMIME):
         :returns: An email without signature and whether the input was signed
         :rtype: (Message,bool)
         """
-        from email.parser import Parser
         from email.message import Message
-        if type(mail)==str: mail = Parser().parsestr(mail)
-        elif not isinstance(mail,Message): return None, False
+        from .mail import protect_mail
+        if type(mail)==str: mail = protect_mail(mail,ending=None)
+        elif not isinstance(mail,Message):
+            raise TypeError("mail must be Message or str")
         payload, signatures, rawmail = self._signature(mail)
         return rawmail, len(signatures)>0
 
@@ -265,6 +273,7 @@ class PGPMIME(KryptoMIME):
         from .mail import protect_mail, fix_lines
         results, key_ids = [], []
         signed, fmt = False, False
+        payload = str(payload) # copy
         for signature in signatures:
             if signature: signature = self._fix_quoting(signature)
             for ending in (None,'\r\n','\n'):
@@ -307,10 +316,10 @@ class PGPMIME(KryptoMIME):
         """
         results = {'encrypted':False,'decryption':None,'signed':False,'key_ids':[],'results':[]}
         # possible encryptions: nothing,only signed, encrypted+signed, encrypted after signed 
-        from email.parser import Parser
         from email.message import Message
         import email.utils
-        if type(mail)==str: mail = Parser().parsestr(mail)
+        from .mail import protect_mail
+        if type(mail)==str: mail = protect_mail(mail,ending=None)
         elif not isinstance(mail,Message): return False, results
         sender = mail.get('from', [])
         sender = self.find_key(email.utils.parseaddr(sender)[1])
@@ -362,18 +371,20 @@ class PGPMIME(KryptoMIME):
         """
         results = {'encrypted':False,'decryption':None,'signed':False,'key_ids':[],'results':[]}
         # possible encryptions: nothing,only signed, encrypted+signed, encrypted after signed 
-        from email.parser import Parser
         from email.message import Message
+        from .mail import protect_mail
         import email.utils
-        if type(mail)==str: mail = Parser().parsestr(mail)
-        elif not isinstance(mail,Message): return None, False, results
+        if type(mail)==str: mail = protect_mail(mail,ending=None)
+        elif not isinstance(mail,Message):
+            raise TypeError("mail must be Message or str")
         sender = mail.get('from', [])
         sender = self.find_key(email.utils.parseaddr(sender)[1])
         ciphertext, is_pgpmime = self._ciphertext(mail)
         if ciphertext: # Ciphertext present? Decode
             results['encrypted'] = True
             if not 'default_key' in kwargs:
-                if not self.default_key: return None, False, results # key required
+                if not self.default_key:
+                    raise KeyMissingError("default")
                 tos = mail.get_all('to', [])
                 ccs = mail.get_all('cc', [])
                 receiver = None
@@ -381,10 +392,10 @@ class PGPMIME(KryptoMIME):
                     receiver = self.find_key(email.utils.formataddr(to))
                     if receiver: break
                 if not receiver:
-                    return None, False, results # cannot decrypt
+                    raise KeyMissingError("receiver")
                 kwargs['default_key'] = receiver
             elif kwargs['default_key']==False or (kwargs['default_key']==True and not self.default_key):
-                return None, False, results # key required
+                raise KeyMissingError("default")
             ciphertext = self._fix_quoting(ciphertext)
             result = self.decrypt_str(ciphertext, **kwargs)
             results['decryption'] = result
@@ -419,18 +430,19 @@ class PGPMIME(KryptoMIME):
         :rtype: (Message,dict)
         """
         from email.message import Message
-        from email.parser import HeaderParser, Parser
         from .mail import protect_mail
-        if not isinstance(mail,(Message,str)): return None, None
-        mail = protect_mail(mail,ending='\r\n',sevenbit=True) # fix line endings + 7bit
+        if not isinstance(mail,(Message,str)):
+            raise TypeError("mail must be Message or str")
+        mail = protect_mail(mail,ending='\r\n',sevenbit=True) # fix line endings + 7bit RFC2822
         if not 'default_key' in kwargs:
             import email.utils
             sender = mail.get('from', [])
             sender = self.find_key(email.utils.parseaddr(sender)[1],secret=True)
-            if not sender: return None, None # key required
+            if not sender:
+                raise KeyMissingError("sender")
             kwargs['default_key'] = sender
         elif kwargs['default_key']==False or (kwargs['default_key']==True and not self.default_key):
-            return None, None # key required
+            raise KeyMissingError("default")
         plaintext, submsg = self._plaintext(mail, inline)
         # Generate signature, report errors
         try:
@@ -491,26 +503,27 @@ class PGPMIME(KryptoMIME):
         """
         import email.utils
         from email.message import Message
-        from email.parser import HeaderParser, Parser
-        from .mail import _mail_addreplace_header
-        if type(mail)==str: mail = Parser().parsestr(mail)
-        elif not isinstance(mail,Message): return None, None
+        from .mail import _mail_addreplace_header, protect_mail
+        if type(mail)==str: mail = protect_mail(mail,ending=None)
+        elif not isinstance(mail,Message):
+            raise TypeError("mail must be Message or str")
         tos = mail.get_all('to', [])
         ccs = mail.get_all('cc', [])
         recipients = [self.find_key(email.utils.formataddr(to)) for to in email.utils.getaddresses(tos + ccs)]
-        if None in recipients: return None, None # key required
+        if None in recipients:
+            raise KeyMissingError("recipients")
         if toself or sign:
             sender = email.utils.parseaddr(mail.get('from', []))[1]
             senderkey = self.find_key(sender,secret=True)
-            if not senderkey: return None, None # key required
+            if not senderkey:
+                raise KeyMissingError("sender")
             if toself:
                 if not senderkey in recipients: recipients.append(senderkey)
         if sign:
             if not 'default_key' in kwargs:
-                #if not self.default_key: return None, None # key required
                 kwargs['default_key'] = senderkey
             elif kwargs['default_key']==False or (kwargs['default_key']==True and not self.default_key):
-                return None, None # key required
+                raise KeyMissingError("default")
         plaintext, submsg = self._plaintext(mail, inline, protect=False)
         # Do encryption, report errors
         kwargs['sign'] = sign
@@ -528,7 +541,7 @@ class PGPMIME(KryptoMIME):
             vresult = self.decrypt_str(ciphertext, **kwargs)
             if not vresult.ok or (sign and not vresult.valid): return None, result
         # Compile encrypted message
-        encmail = HeaderParser().parsestr(mail.as_string())
+        encmail = protect_mail(mail,ending=None)
         _mail_addreplace_header(encmail,'Content-Transfer-Encoding','7bit')
         if not mail.is_multipart() and inline:
             encmail.set_payload(ciphertext)
@@ -660,7 +673,7 @@ class GPGMIME(PGPMIME):
         for line in data.splitlines(True)[3:]: # remove first three lines
             if line.rstrip()=='-----BEGIN PGP SIGNATURE-----': break
             text += line
-        return text
+        return text.rstrip()
 
     def pubkey_attachment(self,key): # pragma: no cover
         "returns an attachment with the specified public key"
